#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2017 by Christian Tremblay, P.Eng <christian.tremblay@servisysDeviceObject.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Device.py - describe a BACnet Device

"""
import asyncio
import logging
import os.path

# --- standard Python modules ---
from collections import namedtuple
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union


# --- this application's modules ---
from bacpypes3.basetypes import ServicesSupported
from bacpypes3.errors import NoResponse

# from ...bokeh.BokehRenderer import BokehPlot
from ...db.sql import SQLMixin
from ...tasks.DoOnce import DoOnce
from ...tasks.Poll import DeviceOneShotPoll
from ..io.IOExceptions import (
    BadDeviceDefinition,
    DeviceNotConnected,
    NoResponseFromController,
    RemovedPointException,
    SegmentationNotSupported,
    WritePropertyException,
    WrongParameter,
)
from ..utils.notes import note_and_log
from ..utils.lookfordependency import pandas_if_available
from .mixins.read_mixin import ReadProperty, ReadPropertyMultiple
from .Points import BooleanPoint, EnumPoint, NumericPoint, OfflinePoint, Point
from .Virtuals import VirtualPoint

_PANDAS, pd, _, _ = pandas_if_available()
# ------------------------------------------------------------------------------


class DeviceProperties(object):
    def __init__(self):
        self.name: str = "Unknown"
        self.address: Optional[str] = None
        self.device_id: Optional[int] = None
        self.network: Optional[Any] = None
        self.pollDelay: Optional[int] = None
        self.objects_list: Optional[List] = None
        self.pss: ServicesSupported = ServicesSupported()
        self.multistates: Optional[Dict] = None
        self.db_name: Optional[str] = None
        self.segmentation_supported: bool = True
        self.history_size: Optional[int] = None
        self.save_resampling: str = "1s"
        self.clear_history_on_save: Optional[bool] = None
        self.bacnet_properties: Dict = {}
        self.auto_save: Optional[bool] = None
        self.fast_polling: bool = False
        self.vendor_id: int = 0
        self.ping_failures: int = 0

    @property
    def asdict(self) -> Dict:
        return self.__dict__

    def __repr__(self):
        return f"{self.asdict}"


@note_and_log
class Device(SQLMixin):
    """
    This class represents a BACnet device. It provides methods to read, write, simulate, and release
    communication with the device on the network.

    Parameters:
    address (str, optional): The address of the device (e.g., '2:5'). Defaults to None.
    device_id (int, optional): The BACnet device ID (boid). Defaults to None.
    network (BAC0.scripts.ReadWriteScript.ReadWriteScript, optional): Defined by BAC0.connect(). Defaults to None.
    poll (int, optional): If greater than 0, the device will poll every point each x seconds. Defaults to None.
    from_backup (str, optional): SQLite backup file. Defaults to None.
    segmentation_supported (bool, optional): When set to False, BAC0 will not use read property multiple to poll the device. Defaults to None.
    object_list (list, optional): User can provide a custom object list for the creation of the device. The object list must be built using the same pattern returned by bacpypes when polling the objectList property. Defaults to None.
    auto_save (bool or int, optional): If False or 0, auto_save is disabled. To activate, pass an integer representing the number of polls before auto_save is called. Will write the histories to SQLite db locally. Defaults to None.
    clear_history_on_save (bool, optional): If set to True, will clear device history. Defaults to None.

    """

    def __init__(
        self,
        address: Optional[str] = None,
        device_id: Optional[int] = None,
        network: Optional[Any] = None,
        *,
        poll: int = 10,
        from_backup: Optional[str] = None,  # filename of backup
        segmentation_supported: bool = True,
        object_list: Optional[List] = None,
        auto_save: bool = False,
        save_resampling: str = "1s",
        clear_history_on_save: bool = False,
        history_size: Optional[int] = None,
        reconnect_on_failure: bool = True,
    ):
        self.properties = DeviceProperties()
        # self.initialized = False
        self.properties.address = address
        self.properties.device_id = device_id
        self.properties.network = network
        self.properties.pollDelay = poll
        self.properties.fast_polling = True if poll < 10 else False
        self.properties.name = ""
        self.properties.vendor_id = 0
        self.properties.objects_list = []
        self.properties.pss = ServicesSupported()
        self.properties.multistates = {}
        self.properties.auto_save = auto_save
        self.properties.save_resampling = save_resampling
        self.properties.clear_history_on_save = clear_history_on_save
        self.properties.history_size = history_size
        self._reconnect_on_failure = reconnect_on_failure

        self.segmentation_supported = segmentation_supported
        self.custom_object_list = object_list

        # self.db = None
        # Todo : find a way to normalize the name of the db
        self.properties.db_name = None

        self.points = []
        self._list_of_trendlogs = {}

        self._polling_task = namedtuple("_polling_task", ["task", "running"])
        self._polling_task.task = None
        self._polling_task.running = False

        self._find_overrides_progress = 0.0
        self._find_overrides_running = False
        self._release_overrides_progress = 0.0
        self._release_overrides_running = False
        self.creation_task = (
            None  # when sent to asyncio.create_tasks, it will be stored here
        )

        self.note("Controller initialized")

        if from_backup:
            filename = from_backup
            db_name = filename.split(".")[0]
            self.properties.network = None
            if os.path.isfile(filename):
                self.properties.db_name = db_name

            else:
                raise FileNotFoundError(f"Can't find {filename} on drive")
        else:
            if (
                self.properties.network
                and self.properties.address
                and self.properties.device_id is not None
            ):
                pass
            else:
                raise BadDeviceDefinition(
                    "Please provide address, device id and network or specify from_backup argument"
                )

    @property
    def initialized(self):
        if isinstance(self, DeviceConnected):
            return True
        return False

    async def new_state(self, newstate: Any) -> None:
        """
        Changes the state of the device.

        This method forms the basis of the state machine mechanism and is used to transition between device states.
        It also calls the state initialization function.

        :param newstate: The new state to transition to.
        :type newstate: Any
        :return: None
        """
        self.log(
            f"Changing {self.properties.device_id} state to {str(newstate).split('.')[-1]}",
            level="info",
        )
        self.__class__ = newstate
        await self._init_state()

    async def _init_state(self) -> None:
        """
        Execute additional code upon state modification
        """
        raise NotImplementedError()

    def connect(self) -> None:
        """
        Connect the device to the network
        """
        raise NotImplementedError()

    def disconnect(self, save_on_disconnect=True, unregister=True) -> None:
        asyncio.create_task(self._disconnect(save_on_disconnect, unregister))

    async def _disconnect(self, save_on_disconnect=True, unregister=True) -> None:
        raise NotImplementedError()

    def initialize_device_from_db(self) -> None:
        raise NotImplementedError()

    def df(self, list_of_points: List[str], force_read: bool = True) -> pd.DataFrame:
        """
        Build a pandas DataFrame from a list of points.  DataFrames are used to present and analyze data.

        :param list_of_points: a list of point names as str
        :returns: pd.DataFrame
        """
        raise NotImplementedError()

    @property
    def simulated_points(self) -> Iterator[Point]:
        """
        iterate over simulated points

        :returns: points if simulated (out_of_service == True)
        :rtype: BAC0.core.devices.Points.Point
        """
        for each in self.points:
            if each.properties.simulated[0]:
                yield each

    async def _buildPointList(self) -> None:
        """
        Read all points from a device into a (Pandas) dataframe (Pandas).  Items are
        accessible by point name.
        """
        raise NotImplementedError()

    def __getitem__(
        self, point_name: Union[str, List[str]]
    ) -> Union[Point, pd.DataFrame]:
        """
        Get a point from its name.
        If a list is passed - a dataframe is returned.

        :param point_name: (str) name of the point or list of point_names
        :type point_name: str
        :returns: (Point) the point (can be Numeric, Boolean or Enum) or pd.DataFrame
        """
        raise NotImplementedError()

    def __iter__(self) -> Point:
        """
        When iterating a device, iterate points of it.
        """
        raise NotImplementedError()

    def __contains__(self, value: str) -> bool:
        "When using in..."
        raise NotImplementedError()

    @property
    def points_name(self) -> List[str]:
        """
        When iterating a device, iterate points of it.
        """
        raise NotImplementedError()

    def to_excel(self) -> None:
        """
        Using xlwings, make a dataframe of all histories and save it
        """
        raise NotImplementedError()

    def __setitem__(self, point_name: str, value: float) -> None:
        """
        Write, sim or ovr value

        :param point_name: Name of the point to set
        :param value: value to write to the point
        :type point_name: str
        :type value: float
        """
        raise NotImplementedError()

    def __len__(self) -> int:
        """
        Will return number of points available
        """
        raise NotImplementedError()

    def _parseArgs(self, arg: str) -> Tuple[str, str]:
        """
        Given a string, interpret the last word as the value, everything else is
        considered to be the point name.
        """
        args = arg.split()
        pointName = " ".join(args[:-1])
        value = args[-1]
        return (pointName, value)

    def clear_histories(self) -> None:
        for point in self.points:
            point.clear_history()

    def update_history_size(self, size: Optional[int] = None) -> None:
        for point in self.points:
            point.properties.history_size = size

    @property
    def analog_units(self) -> Dict[str, str]:
        raise NotImplementedError()

    @property
    def temperatures(self) -> Dict[str, str]:
        raise NotImplementedError()

    @property
    def percent(self) -> Dict[str, str]:
        raise NotImplementedError()

    @property
    def multi_states(self) -> Dict[str, str]:
        raise NotImplementedError()

    @property
    def binary_states(self) -> Dict[str, str]:
        raise NotImplementedError()

    def _findPoint(self, name: str, force_read: bool = True) -> Point:
        """
        Helper that retrieve point based on its name.

        :param name: (str) name of the point
        :param force_read: (bool) read value of the point each time the function is called.
        :returns: Point object
        :rtype: BAC0.core.devices.Point.Point (NumericPoint, EnumPoint or BooleanPoint)
        """
        raise NotImplementedError()

    def find_point(self, objectType: str, objectAddress: float) -> Point:
        """
        Find point based on type and address
        """
        for point in self.points:
            if (
                point.properties.type == objectType
                and float(point.properties.address) == objectAddress
            ):
                return point
        raise ValueError(f"{objectType} {objectAddress} doesn't exist in controller")

    def find_overrides(self, force: bool = False) -> None:
        if self._find_overrides_running and not force:
            self.log(
                f"Already running ({self._find_overrides_progress:.1%})... please wait.",
                level="warning",
            )
            return
        lst = []
        self._find_overrides_progress = 0.0
        self._find_overrides_running = True
        total = len(self.points)

        def _find_overrides() -> None:
            self._log.warning(
                "Overrides are being checked, wait for completion message."
            )
            for idx, point in enumerate(self.points):
                if point.is_overridden:
                    lst.append(point)
                self._find_overrides_progress = idx / total
            self._log.warning(
                "Override check ready, results available in device.properties.points_overridden"
            )
            self.properties.points_overridden = lst
            self._find_overrides_running = False
            self._find_overrides_progress = 1.0

        self.do(_find_overrides)

    def find_overrides_progress(self) -> float:
        return self._find_overrides_progress

    def release_all_overrides(self, force: bool = False) -> None:
        if self._release_overrides_running and not force:
            self.log(
                f"Already running ({self._release_overrides_progress:.1%})... please wait.",
                level="warning",
            )
            return
        self._release_overrides_running = True
        self._release_overrides_progress = 0.0

        def _release_all_overrides() -> None:
            self.find_overrides()
            while self._find_overrides_running:
                self._release_overrides_progress = self._find_overrides_progress * 0.5

            if self.properties.points_overridden:
                total = len(self.properties.points_overridden)
                self.log("=================================", level="info")
                self.log("Overrides found... releasing them", level="info")
                self.log("=================================", level="info")
                for idx, point in enumerate(self.properties.points_overridden):
                    self.log(f"Releasing {point}", level="info")
                    point.release_ovr()
                    self._release_overrides_progress = (idx / total) / 2 + 0.5
            else:
                self.log("No override found", level="info")

            self._release_overrides_running = False
            self._release_overrides_progress = 1

        self.do(_release_all_overrides)

    def do(self, func: Any) -> None:
        DoOnce(func).start()

    def __repr__(self) -> str:
        return f"{self.properties.name} / Undefined"


# def device(*args: Any, **kwargs: Any) -> Device:
#    dev = Device(*args, **kwargs)
#    t = asyncio.create_task(dev.new_state(DeviceDisconnected))
#    dev.creation_task = t
#    while not t.done:
#        pass
#    return dev


async def device(*args: Any, **kwargs: Any) -> Device:
    dev = Device(*args, **kwargs)
    await dev.new_state(DeviceDisconnected)
    return dev


# @fix_docs
class DeviceConnected(Device):
    """
    Find a device on the BACnet network.  Set its state to 'connected'.
    Once connected, all subsequent commands use this BACnet connection.
    """

    async def _init_state(self):
        await self._buildPointList()
        self.properties.network.register_device(self)
        if self.properties.pollDelay == 0:
            _poll = DeviceOneShotPoll(self)
            _poll.start()
        # self.initialized = True

    async def _disconnect(self, save_on_disconnect=True, unregister=True):
        self.log(
            f"Wait while stopping polling for {self.properties.name}", level="info"
        )
        self.poll(command="stop")
        if unregister:
            self.properties.network.unregister_device(self)
            self.properties.network = None
        if save_on_disconnect:
            self.log(f"Saving {self.properties.name} to database...", level="info")
            await self.save()
        if self.properties.db_name:
            await self.new_state(DeviceFromDB)
        else:
            await self.new_state(DeviceDisconnected)

    async def connect(self, *, db=None):
        """
        A connected device can be switched to 'database mode' where the device will
        not use the BACnet network but instead obtain its contents from a previously
        stored database.
        """
        if db:
            self.poll(command="stop")
            self.properties.db_name = db.split(".")[0]
            await self.new_state(DeviceFromDB)
        else:
            self._log.warning(
                "Already connected, provide db arg if you want to connect to db"
            )

    def df(self, list_of_points, force_read=True):
        """
        When connected, calling DF should force a reading on the network.
        """

        his = []
        for point in list_of_points:
            try:
                his.append(self._findPoint(point, force_read=force_read).history)
            except ValueError as ve:
                self.log(f"Value Error : {ve}", level=logging.DEBUG)
                continue
        if not _PANDAS:
            return dict(zip(list_of_points, his))
        return pd.DataFrame(dict(zip(list_of_points, his)))

    async def _buildPointList(self):
        """
        Upon connection to build the device point list and properties.
        """
        try:
            self.properties.pss.value = await self.properties.network.read(
                "{} device {} protocolServicesSupported".format(
                    self.properties.address, self.properties.device_id
                )
            )

        except NoResponseFromController as error:
            self.log(f"Controller not found, aborting. ({error})", level="error")
            return ("Not Found", "", [], [])

        except SegmentationNotSupported:
            self.log("Segmentation not supported", level="warning")
            self.segmentation_supported = False
            await self.new_state(DeviceDisconnected)

        self.properties.name = str(
            await self.properties.network.read(
                f"{self.properties.address} device {self.properties.device_id} objectName"
            )
        )
        self.properties.vendor_id = await self.properties.network.read(
            "{} device {} vendorIdentifier".format(
                self.properties.address, self.properties.device_id
            )
        )
        self._log.info(
            "Device {}:[{}] found... building points list".format(
                self.properties.device_id, self.properties.name
            )
        )
        try:
            (
                self.properties.objects_list,
                self.points,
                self._list_of_trendlogs,
            ) = await self._discoverPoints(self.custom_object_list)
            if self.properties.pollDelay is not None and self.properties.pollDelay > 0:
                self.poll(delay=self.properties.pollDelay)
            self.update_history_size(size=self.properties.history_size)
            self._log.info(
                f"Device {self.properties.name} | {self.properties.device_id} ready, use device_name.points and start interact with it"
            )
            # self.clear_histories()
        except NoResponseFromController:
            self.log("Cannot retrieve object list, disconnecting...", level="error")
            self.segmentation_supported = False
            await self.new_state(DeviceDisconnected)
        except IndexError:
            if self._reconnect_on_failure:
                self.log("Device creation failed... re-connecting", level="error")
                await self.new_state(DeviceDisconnected)
            else:
                self.log("Device creation failed... disconnecting", level="error")

    def __getitem__(self, point_name):
        """
        Allows the syntax: device['point_name'] or device[list_of_points]

        If calling a list, last value will be used (won't read on the network)
        for performance reasons.
        If calling a simple point, point will be read via BACnet.
        """
        try:
            if isinstance(point_name, list):
                return self.df(point_name, force_read=False)
            elif isinstance(point_name, tuple):
                _type, _address = point_name
                for point in self.points:
                    if point.properties.type == _type and str(
                        point.properties.address
                    ) == str(_address):
                        return point
            else:
                try:
                    return self._findPoint(point_name, force_read=False)
                except ValueError:
                    try:
                        return self._findTrend(point_name)
                    except ValueError:
                        try:
                            if "@prop_" in point_name:
                                point_name = point_name.split("prop_")[1]
                                return self.read_property(
                                    ("device", self.properties.device_id, point_name)
                                )
                            else:
                                raise ValueError()
                        except ValueError:
                            raise ValueError()
        except ValueError as ve:
            self.log(f"Can't find {point_name} | {ve}", level="error")

    def __iter__(self):
        yield from self.points

    def __contains__(self, value):
        """
        Allows the syntax:
            if "point_name" in device:
        """
        return value in self.points_name

    @property
    def pollable_points_name(self):
        for each in self.points:
            if not isinstance(each, VirtualPoint):
                yield each.properties.name
            else:
                continue

    @property
    def points_name(self):
        for each in self.points:
            yield each.properties.name

    def __setitem__(self, point_name, value):
        """
        Allows the syntax:
            device['point_name'] = value
        """
        try:
            loop = asyncio.get_event_loop()
            loop.create_task(self._findPoint(point_name)._set(value))
        except WritePropertyException as ve:
            self.log(f"{ve}", level="error")

    def __len__(self):
        """
        Length of a device = number of points
        """
        return len(self.points)

    def _parseArgs(self, arg):
        args = arg.split()
        pointName = " ".join(args[:-1])
        value = args[-1]
        return (pointName, value)

    @property
    def analog_units(self):
        """
        Shortcut to retrieve all analog points units [Used by Bokeh trending feature]
        """
        au = []
        us = []
        for each in self.points:
            if isinstance(each, NumericPoint):
                au.append(each.properties.name)
                us.append(each.properties.units_state)
        return dict(zip(au, us))

    @property
    def temperatures(self):
        for each in self.analog_units.items():
            if "deg" in each[1]:
                yield each

    @property
    def percent(self):
        for each in self.analog_units.items():
            if "percent" in each[1]:
                yield each

    @property
    def multi_states(self):
        ms = []
        us = []
        for each in self.points:
            if isinstance(each, EnumPoint):
                ms.append(each.properties.name)
                us.append(each.properties.units_state)
        return dict(zip(ms, us))

    @property
    def binary_states(self):
        bs = []
        us = []

        for each in self.points:
            if isinstance(each, BooleanPoint):
                bs.append(each.properties.name)
                us.append(each.properties.units_state)
        return dict(zip(bs, us))

    def _findPoint(self, name, force_read=False):
        """
        Used by getter and setter functions
        """
        for point in self.points:
            if point.properties.name == name:
                if force_read:
                    point.value
                return point
        raise ValueError(f"{name} doesn't exist in controller")

    def _trendlogs(self):
        for k, v in self._list_of_trendlogs.items():
            name, trendlog = v
            yield trendlog

    @property
    def trendlogs_names(self):
        for each in self._trendlogs():
            yield each.properties.object_name

    @property
    def trendlogs(self):
        return list(self._trendlogs())

    def _findTrend(self, name):
        for trend in self._trendlogs():
            if trend.properties.object_name == name:
                return trend
        raise ValueError(f"{name} doesn't exist in controller")

    async def read_property(self, prop):
        # if instance == -1:
        #    pass
        if isinstance(prop, tuple):
            _obj, _instance, _prop = prop
        elif isinstance(prop, str):
            _obj = "device"
            _instance = self.properties.device_id
            _prop = prop
        else:
            raise ValueError(
                "Please provide property using tuple with object, instance and property"
            )
        try:
            request = f"{self.properties.address} {_obj} {_instance} {_prop}"
            val = await self.properties.network.read(
                request, vendor_id=self.properties.vendor_id
            )
        except KeyError as error:
            raise Exception(f"Unknown property : {error}")
        return val

    async def write_property(self, prop, value, priority=None):
        if prop == "description":
            self.update_description(value)
        else:
            if priority is not None:
                priority = f"- {priority}"
            if isinstance(prop, tuple):
                _obj, _instance, _prop = prop
            else:
                raise ValueError(
                    "Please provide property using tuple with object, instance and property"
                )
            try:
                request = "{} {} {} {} {} {}".format(
                    self.properties.address, _obj, _instance, _prop, value, priority
                )
                val = await self.properties.network.write(
                    request, vendor_id=self.properties.vendor_id
                )
            except KeyError as error:
                raise Exception(f"Unknown property : {error}")
            return val

    async def update_bacnet_properties(self):
        """
        Retrieve bacnet properties for this device

        """
        try:
            res = await self.properties.network.readMultiple(
                "{} device {} all".format(
                    self.properties.address, str(self.properties.device_id)
                ),
                vendor_id=self.properties.vendor_id,
                show_property_name=True,
            )
            for each in res:
                if not each:
                    continue
                v, prop = each
                self.properties.bacnet_properties[prop] = v

        except Exception as e:
            raise Exception(f"Problem reading : {self.properties.name} | {e}")

    def _bacnet_properties(self, update=False):
        if not self.properties.bacnet_properties or update:
            self.update_bacnet_properties()
        return self.properties.bacnet_properties

    @property
    def bacnet_properties(self):
        return self._bacnet_properties(update=True)

    async def update_description(self, value):
        await self.properties.network.send_text_write_request(
            addr=self.properties.address,
            obj_type="device",
            obj_inst=int(self.device_id),
            value=value,
            prop_id="description",
        )
        self.properties.description = str(await self.read_property("description"))

    async def ping(self):
        try:
            if await self.read_property("objectName") == self.properties.name:
                self.properties.ping_failures = 0
                return True
            else:
                self.properties.ping_failures += 1
                return False
        except NoResponseFromController as e:
            self.log(
                f"{self.properties.name} ({self.properties.address})| Ping failure ({e}).",
                level="error",
            )

            self.properties.ping_failures += 1
            return False

    def __repr__(self):
        return f"{self.properties.name} / Connected"


# ------------------------------------------------------------------------------


class RPDeviceConnected(DeviceConnected, ReadProperty):
    """
    [Device state] If device is connected but doesn't support ReadPropertyMultiple

    BAC0 will not poll such points automatically (since it would cause excessive network traffic).
    Instead manual polling must be used as needed via the poll() function.
    """

    def __str__(self):
        return "connected [for ReadProperty]"


class RPMDeviceConnected(DeviceConnected, ReadPropertyMultiple):
    """
    [Device state] If device is connected and supports ReadPropertyMultiple
    """

    def __str__(self):
        return "connected [for ReadPropertyMultiple]"


# @fix_docs
class DeviceDisconnected(Device):
    """
    [Device state] Initial state of a device. Disconnected from BACnet.
    """

    async def _init_state(self):
        # self.initialized = False
        await self.connect()

    async def connect(self, *, db=None, network=None):
        """
        Attempt to connect to device.  If unable, attempt to connect to a controller database
        (so the user can use previously saved data).
        """
        if network:
            self.properties.network = network
        if not self.properties.network:
            self.log("No network...calling DeviceFromDB", level="debug")
            if db:
                await self.new_state(DeviceFromDB)
            self.log(
                'You can reconnect to network using : "device.connect(network=bacnet)"',
                level="info",
            )

        else:
            try:
                await self.properties.network.read(
                    "{} device {} objectName".format(
                        self.properties.address, self.properties.device_id
                    )
                )

                segmentation = await self.properties.network.read(
                    "{} device {} segmentationSupported".format(
                        self.properties.address, self.properties.device_id
                    )
                )

                self.segmentation_supported = (
                    False if segmentation and segmentation.numerator == 3 else True
                )

                if self.segmentation_supported:
                    await self.new_state(RPMDeviceConnected)
                else:
                    await self.new_state(RPDeviceConnected)

            except SegmentationNotSupported:
                self.segmentation_supported = False
                self._log.warning(
                    "Segmentation not supported.... expect slow responses."
                )
                await self.new_state(RPDeviceConnected)

            except (NoResponseFromController, AttributeError) as error:
                self._log.warning(f"BAC0 got no response from controller: {error}")
                if self.properties.db_name:
                    await self.new_state(DeviceFromDB)
                else:
                    self._log.warning(
                        "Offline: provide database name to load stored data."
                    )
                    self._log.warning("Ex. controller.connect(db = 'backup')")

    def df(self, list_of_points, force_read=True):
        raise DeviceNotConnected("Must connect to BACnet or database")

    @property
    def simulated_points(self):
        for each in self.points:
            if each.properties.simulated:
                yield each

    def _buildPointList(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    # This should be a "read" function and rpm defined in state rpm
    def read_multiple(
        self, points_list, *, points_per_request=25, discover_request=(None, 6)
    ):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def poll(self, command="start", *, delay=10):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __getitem__(self, point_name):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __iter__(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __contains__(self, value):
        raise DeviceNotConnected("Must connect to BACnet or database")

    @property
    def points_name(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def to_excel(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __setitem__(self, point_name, value):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __len__(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    @property
    def analog_units(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    @property
    def temperatures(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    @property
    def percent(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    @property
    def multi_states(self):
        raise DeviceNotConnected("Must connect to bacnet or database")

    @property
    def binary_states(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def _discoverPoints(self, custom_object_list=None):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def _findPoint(self, name, force_read=True):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __repr__(self):
        return f"{self.properties.name} / Disconnected"


# ------------------------------------------------------------------------------

# @fix_docs


class DeviceFromDB(DeviceConnected):
    """
    [Device state] Where requests for a point's present value returns the last
    valid value from the point's history.
    """

    async def _init_state(self):
        try:
            await self.initialize_device_from_db()
        except ValueError as e:
            self.log(f"Problem with DB initialization : {e}", level="error")
            # self.new_state(DeviceDisconnected)
            raise

    async def connect(self, *, network=None, from_backup=None):
        """
        In DBState, a device can be reconnected to BACnet using:
            device.connect(network=bacnet) (bacnet = BAC0.connect())
        """
        if network and from_backup:
            raise WrongParameter("Please provide network OR from_backup")

        elif network:
            self.log("Network provided... trying to connect", level="debug")
            self.properties.network = network
            try:
                name = await self.properties.network.read(
                    "{} device {} objectName".format(
                        self.properties.address, self.properties.device_id
                    )
                )

                segmentation = await self.properties.network.read(
                    "{} device {} segmentationSupported".format(
                        self.properties.address, self.properties.device_id
                    )
                )
                segmentation_supported = False if segmentation.numerator == 3 else True

                if name:
                    if segmentation_supported:
                        self.log("Segmentation supported, connecting...", level="debug")
                        await self.new_state(RPMDeviceConnected)
                    else:
                        self.log(
                            "Segmentation not supported, connecting...", level="debug"
                        )
                        await self.new_state(RPDeviceConnected)
                    # self.db.close()

            except (NoResponseFromController, NoResponse):
                self.log("Unable to connect, keeping DB mode active", level="error")

        else:
            self.log("Not connected, open DB", level="debug")
            if from_backup:
                self.properties.db_name = from_backup.split(".")[0]
            await self._init_state()

    async def initialize_device_from_db(self):
        self.log(f"Initializing DB for {self.properties.name}", level="info")
        # Save important properties for reuse
        if self.properties.db_name:
            dbname = self.properties.db_name
            try:
                self._props = self.read_dev_prop(self.properties.db_name)
            except ValueError:
                raise ValueError(f"Can't find {self.properties.db_name} on drive")
        else:
            self.log(f"Missing argument DB for {self.properties.name}", level="info")
            raise ValueError(
                f"Please provide db name using device.load_db('name') for {self.properties.name}"
            )

        # network = self.properties.network
        pss = self.properties.pss

        self.points = []
        for point in await self.points_from_sql(self.properties.db_name):
            try:
                self.points.append(OfflinePoint(self, point))
            except RemovedPointException:
                continue

        self.properties = DeviceProperties()
        self.properties.db_name = dbname
        self.properties.address = self._props["address"]
        self.properties.device_id = self._props["device_id"]
        self.properties.network = None
        self.properties.pollDelay = self._props["pollDelay"]
        self.properties.name = self._props["name"]
        self.properties.objects_list = self._props["objects_list"]
        self.properties.pss = pss
        self.properties.serving_chart = {}
        self.properties.charts = []
        self.properties.multistates = self._props["multistates"]
        self.properties.auto_save = self._props["auto_save"]
        self.properties.save_resampling = self._props["save_resampling"]
        self.properties.clear_history_on_save = self._props["clear_history_on_save"]
        self.properties.default_history_size = self._props["history_size"]
        self.log(f"{self.properties.name} restored from db", level="info")
        self.log(
            'You can reconnect to network using : "device.connect(network=bacnet)"',
            level="info",
        )

    @property
    def simulated_points(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def _buildPointList(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    # This should be a "read" function and rpm defined in state rpm
    def read_multiple(
        self, points_list, *, points_per_request=25, discover_request=(None, 6)
    ):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def poll(self, command="start", *, delay=10):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __contains__(self, value):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def to_excel(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __setitem__(self, point_name, value):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def _discoverPoints(self, custom_object_list=None):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __repr__(self):
        return f"{self.properties.name} / Disconnected"


# ------------------------------------------------------------------------------


class DeviceLoad(DeviceFromDB):
    def __init__(self, filename=None):
        if filename:
            Device.__init__(self, None, None, None, from_backup=filename)
        else:
            raise Exception("Please provide backup file as argument")



#!/usr/bin/env python
# -*- coding utf-8 -*-

"""
Test Bacnet communication with another device
"""
import asyncio
import pytest

NEWCSVALUE = "New_Test"


@pytest.mark.asyncio
async def test_WriteAV(network_and_devices):
    async for resources in network_and_devices:
        loop, bacnet, device_app, device30_app, test_device, test_device_30 = resources
        # Write to an object and validate new value is correct
        old_value = await test_device["AV"].value
        test_device["AV"] = 11.2
        await asyncio.sleep(1.5)  # or cache will play a trick on you
        new_value = await test_device["AV"].value
        assert (new_value - 11.2) < 0.01


@pytest.mark.asyncio
async def test_RelinquishDefault(network_and_devices):
    async for resources in network_and_devices:
        loop, bacnet, device_app, device30_app, test_device, test_device_30 = resources
        test_device = test_device
        # Write to an object and validate new value is correct
        old_value = await test_device["AV"].value
        test_device["AV"].default(90)
        # time.sleep(1)
        new_value = await test_device["AV"].value
        assert (new_value - 90) < 0.01


@pytest.mark.asyncio
async def test_WriteCharStr(network_and_devices):
    # Write to an object and validate new value is correct
    async for resources in network_and_devices:
        loop, bacnet, device_app, device30_app, test_device, test_device_30 = resources
        test_device = test_device
        test_device["CS_VALUE"] = NEWCSVALUE
        # time.sleep(1)
        new_value = await test_device["CS_VALUE"].value
        assert new_value == NEWCSVALUE


@pytest.mark.skip(
    "Not ready yet as BAC0 do not support out_of_service write -> unlocking PV"
)
async def test_SimulateAI(network_and_devices):
    # Write to an object and validate new value is correct
    async for resources in network_and_devices:
        loop, bacnet, device_app, device30_app, test_device, test_device_30 = resources
        test_device["AI"] = 1
        # time.sleep(1)
        new_value = test_device["AI"].value
        assert test_device.read_property(("analogInput", 0, "outOfService"))
        # something is missing so pv can be written to if outOfService == True
        # assert new_value == 1


@pytest.mark.skip(
    "Not ready yet as BAC0 do not support out_of_service write -> unlocking PV"
)
async def test_RevertSimulation(network_and_devices):
    # Write to an object and validate new value is correct
    async for resources in network_and_devices:
        loop, bacnet, device_app, device30_app, test_device, test_device_30 = resources
        test_device["AI"] = "auto"
        # time.sleep(1)
        new_value = test_device["AI"].value
        assert not test_device.read_property(("analogInput", 0, "outOfService"))
        assert (new_value - 99.9) < 0.01



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
TaskManager.py - creation of threads used for repetitive tasks.

A key building block for point simulation.
"""
import asyncio
import time
from random import random

# --- 3rd party modules ---
# --- this application's modules ---
from ..core.utils.notes import note_and_log

# ------------------------------------------------------------------------------


async def stopAllTasks():
    Task._log.info("Stopping all tasks")
    for each in Task.tasks:
        each.aio_task.cancel()
    Task._log.info("Ok all tasks stopped")
    Task.clean_tasklist(all=True)
    return True


@note_and_log
class Task(object):
    tasks = []
    high_latency = 60

    @classmethod
    def clean_tasklist(cls, all=False):
        if all is True:
            cls._log.debug("Cleaning tasks list")
            cls.tasks = []
        else:
            for each in cls.tasks:
                if each.done:
                    cls._log.debug(f"Removing task {each.name}")
                    cls.tasks.remove(each)

    @classmethod
    def number_of_tasks(cls):
        return len(cls.tasks)

    def __init__(self, fn=None, name=None, delay=0):
        # delay = 0 -> one shot
        self.id = id(self)
        self.name = name if name is not None else f"Task_{self.id}"
        if isinstance(fn, tuple):
            self.fn, self.args = fn
        else:
            self.fn = fn
            self.args = None

        if delay > 0:
            self.delay = delay if delay >= 5 else 5
        else:
            self.delay = 0
        self.previous_execution = None
        self.average_execution_delay = 0
        self.average_latency = 0
        self.next_execution = time.time() + delay + (random() * 10)
        self.execution_time = 0.0
        self.count = 0

        self._kwargs = None
        self._task = None
        self.aio_task = None

    async def task(self):
        raise NotImplementedError("Must be implemented")

    async def execute(self):
        if self.delay > 0:
            self.log(
                f"Installing recurring task {self.name} (id:{self.id})", level="info"
            )
            while True:
                self.count += 1
                _start_time = time.time()
                self.log(
                    f"Executing : {self.name} | Count : {self.count}", level="debug"
                )
                self.log(f"Start Time : {_start_time}", level="debug")
                if self.previous_execution:
                    self.log(
                        f"Previous execution : {self.previous_execution}", level="debug"
                    )
                else:
                    self.log("First Run", level="debug")

                self.average_latency = (
                    self.average_latency + (_start_time - self.next_execution)
                ) / 2
                try:
                    if self.fn and self.args is not None:
                        await self.fn(self.args)
                    elif self.fn:
                        await self.fn()
                    else:
                        if self._kwargs is not None:
                            await self.task(**self._kwargs)
                        else:
                            await self.task()
                except Exception as error:
                    self.log(
                        f"An exception occured while running the task {self.name} (id:{self.id}) : {error}",
                        level="error",
                    )
                if self.previous_execution:
                    _total = self.average_execution_delay + (
                        _start_time - self.previous_execution
                    )
                    self.average_execution_delay = _total / 2
                else:
                    self.average_execution_delay = self.delay

                # self.log('Stat for task {}'.format(self), level='info')
                if self.average_latency > Task.high_latency:
                    self.log(f"High latency for {self.name}", level="warning")
                    self.log(f"Stats : {self}", level="warning")

                self.execution_time = time.time() - _start_time
                self.log(f"Execution Time : {self.execution_time}", level="debug")
                self.previous_execution = _start_time
                self.next_execution = time.time() + self.delay
                await asyncio.sleep(self.delay)
        else:  # one shot
            self.log(f"Running one shot task {self.name} (id:{self.id})", level="info")
            if self.fn and self.args is not None:
                await self.fn(self.args)
            elif self.fn:
                await self.fn()
            else:
                if self._kwargs is not None:
                    await self.task(**self._kwargs)
                else:
                    await self.task()

    def start(self):
        self.aio_task = asyncio.create_task(self.execute(), name=f"aio{self.name}")
        Task.tasks.append(self)

    def stop(self):
        for each in Task.tasks:
            if each.id == self.id:
                each.aio_task.cancel()
                Task.tasks.remove(each)
                return True

    @property
    def done(self):
        if self.aio_task is not None:
            return self.aio_task.done()
        else:
            return False

    @property
    def last_time(self):
        return time.strftime(
            "%Y-%m-%d %H:%M:%S", time.localtime(self.previous_execution)
        )

    @property
    def next_time(self):
        return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.next_execution))

    @property
    def latency(self):
        return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.average_latency))

    def is_alive(self):
        return

    def __repr__(self):
        return "{:<40} | Avg exec delay : {:.2f} sec | Avg latency : {:.2f} sec | last executed : {} | Next Time : {}".format(
            self.name,
            self.average_execution_delay,
            self.average_latency,
            self.last_time,
            self.next_time,
        )

    def __lt__(self, other):
        # list sort use __lt__... little cheat to reverse list already
        return self.next_execution > other.next_execution

    def __eq__(self, other):
        # list remove use __eq__... so compare with id
        if isinstance(other, Task):
            return self.id == other.id
        else:
            return self.id == other


@note_and_log
class OneShotTask(Task):
    def __init__(self, fn=None, args=None, name="Oneshot"):
        if args is not None:
            _args = (fn, args)
        else:
            _args = fn
        super().__init__(fn=_args, name=name, delay=0)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
# --- standard Python modules ---
import datetime
import typing as t

from bacpypes3.apdu import WritePropertyRequest
from bacpypes3.app import Application
from bacpypes3.basetypes import CalendarEntry, DateRange
from bacpypes3.constructeddata import Any, ArrayOf

# --- 3rd party modules ---
from bacpypes3.pdu import Address

from ...core.app.asyncApp import BAC0Application
from ...core.utils.notes import note_and_log


@note_and_log
class Calendar:
    """
    Everything you need to write dates and date ranges to a calendar object.
    """

    DateList = ArrayOf(CalendarEntry)
    datelist_example = {
        "dates": [
            {"date": "2021/3/14", "recurring": False},
            {"date": "2021/3/10", "recurring": True},
        ],
        "dateRanges": [
            {"startDate": "2021/3/16", "endDate": "2021/3/21"},
            {"startDate": "2021/3/5", "endDate": "2021/3/7"},
        ],
    }

    def create_calendar(self, dict_calendar):
        """
        From a structured dict (see dateList_example), create a DateList
        an ArrayOf(CalendarEntry)
        """

        entries = []
        if "dates" in dict_calendar.keys():
            for date_entry in dict_calendar["dates"]:
                year, month, day = (int(x) for x in date_entry["date"].split("/"))
                if date_entry["recurring"]:
                    weekday = 255
                else:
                    weekday = datetime.date(year, month, day).weekday() + 1
                    if weekday > 7:
                        weekday = 1
                _date = (year - 1900, month, day, weekday)
                entries.append(CalendarEntry(date=_date))

        if "dateRanges" in dict_calendar.keys():
            for date_range_entry in dict_calendar["dateRanges"]:
                year, month, day = (
                    int(x) for x in date_range_entry["startDate"].split("/")
                )
                weekday = datetime.date(year, month, day).weekday() + 1
                if weekday > 7:
                    weekday = 1
                start_date = (year - 1900, month, day, weekday)

                year, month, day = (
                    int(x) for x in date_range_entry["endDate"].split("/")
                )
                weekday = datetime.date(year, month, day).weekday() + 1
                if weekday > 7:
                    weekday = 1
                end_date = (year - 1900, month, day, weekday)

                date_range = DateRange(startDate=start_date, endDate=end_date)
                entries.append(CalendarEntry(dateRange=date_range))

        dateList = self.DateList(entries)
        return dateList

    def make_calendar_request(self, destination, object_instance, dateList):
        request = WritePropertyRequest(
            objectIdentifier=("calendar", object_instance),
            propertyIdentifier="dateList",
        )

        address = Address(destination)
        request.pduDestination = address
        request.propertyValue = Any(None)
        request.propertyValue.cast_in(dateList)
        return request

    def send_calendar_request(self, request, timeout=10):
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app

        self.log(f"{'- request:':>12} {request}", level="debug")

        _app.request(request)

        self._log.info(
            f"Calendar Write request sent to device : {request.pduDestination}"
        )

    def write_calendar_dateList(self, destination, calendar_instance, dates):
        calendar = self.create_calendar(dates)
        request = self.make_calendar_request(
            destination=destination,
            object_instance=calendar_instance,
            dateList=calendar,
        )
        self.send_calendar_request(request)

    async def read_calendar_dateList(self, address, calendar_instance):
        """
        This function will read the dateList property of given calendar object and
        pass it to decode_dateList() to convert it into a human readable dict.
        """

        try:
            dateList_object = await self.read(
                f"{address} calendar {calendar_instance} dateList"
            )
            dict_calendar = self.decode_dateList(dateList_object)
        except Exception as error:
            self.log(f"exception: {error!r}", level="error")
            return {}

        return dict_calendar

    def decode_dateList(self, dateList_object) -> t.Dict[str, t.List[t.Dict]]:
        dict_calendar = {"dates": [], "dateRanges": []}
        for entry in dateList_object:
            entry_dict: t.Dict[str, t.Union[str, bool]] = {}
            if entry.date:
                if entry.date[3] == 255:
                    recurring = True
                else:
                    recurring = False
                entry_dict["date"] = "{}/{}/{}".format(
                    entry.date[0] + 1900, entry.date[1], entry.date[2]
                )
                entry_dict["recurring"] = recurring
                dict_calendar["dates"].append(entry_dict)
            elif entry.dateRange:
                entry_dict["startDate"] = "{}/{}/{}".format(
                    entry.dateRange.startDate[0] + 1900,
                    entry.dateRange.startDate[1],
                    entry.dateRange.startDate[2],
                )
                entry_dict["endDate"] = "{}/{}/{}".format(
                    entry.dateRange.endDate[0] + 1900,
                    entry.dateRange.endDate[1],
                    entry.dateRange.endDate[2],
                )
                dict_calendar["dateRanges"].append(entry_dict)

        return dict_calendar



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
# from . import Base, Lite



from collections import namedtuple
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union

from bacpypes3.basetypes import (
    Date,
    DateTime,
    EventState,
    LoggingType,
    LogRecord,
    LogRecordLogDatum,
    Reliability,
    StatusFlags,
    Time,
)
from bacpypes3.constructeddata import ListOf
from bacpypes3.primitivedata import Unsigned
from ...utils.lookfordependency import pandas_if_available

PANDAS, _, _, _ = pandas_if_available()


Record = namedtuple(
    "Record", "timestamp value statusFlags sequencenumber interval trendFlag logEvent"
)


class LocalTrendLog(object):
    """
    Local trendLogs require a databse between values read on the field
    and values used to create thje local trendLogs object.

    I will use the existing SQL code I have to store the local values

    """

    def __init__(self, obj: Any, datatype: str):
        self.obj = obj
        self.data: List[Record] = []
        self.bufferSize: int = 250
        self.statusFlags = StatusFlags([0, 0, 0, 0])
        self.datatype = datatype

    @staticmethod
    def to_float(val: Union[int, float, str]) -> Optional[float]:
        try:
            return float(val)
        except ValueError:
            return None

    @staticmethod
    def decompose_datetime(dt: datetime) -> Tuple[int, int, int, int, int]:
        y = dt.year
        M = dt.month
        d = dt.day
        h = dt.hour
        m = dt.minute
        s = dt.second
        ms = dt.microsecond
        wk = dt.weekday()
        return (y, M, d, wk, h, m, s, ms)

    def to_bacpypes_datetime(self, dt: datetime) -> DateTime:
        _y, _M, _d, wk, _h, _m, _s, _ms = self.decompose_datetime(dt)
        try:
            result = DateTime(date=Date((_y, _M, _d, wk)), time=Time((_h, _m, _s, _ms)))
        except TypeError:
            raise TypeError(f"Error with {dt} {_y=}, {_M=}, {_d=}, {_h=}, {_m=}")
        return result

    def to_logDatum(self, value: Union[int, float, str]) -> Dict[str, Any]:
        _klass = getattr(LogRecordLogDatum, self.datatype)
        return {self.datatype: _klass(value)}

    def to_bacpypes_logrecord(self, record: Record) -> LogRecord:
        """
        For now, only support real... make it work first
        """
        _timestamp = record.timestamp.astimezone()
        _dt = self.to_bacpypes_datetime(_timestamp)
        record_value = self.to_logDatum(record.value)

        return LogRecord(
            timestamp=_dt,
            logDatum=LogRecordLogDatum(**record_value),
            statusFlags=record.statusFlags,
        )

    def add_data(
        self,
        timestamp: datetime,
        value: Union[int, float, str],
        flags: StatusFlags = StatusFlags([0, 0, 0, 0]),
        interval: Optional[int] = None,
        update_after: bool = True,
    ) -> None:
        """
        each object will contain a dict of values that will be
        turned into log_record.
        """
        if not self.data or self.data[-1] == (2**32) - 1:
            sequencenumber = 1
        else:
            sequencenumber = self.data[-1].sequencenumber + 1
        _rec = Record(
            timestamp,
            value,
            flags,
            sequencenumber,
            interval,
            trendFlag=None,
            logEvent=None,
        )
        if _rec.timestamp not in [each.timestamp for each in self.data]:
            self.data.append(_rec)

        # limit to 250 values
        self.data = self.data[-self.bufferSize :]
        if update_after:
            self.update_properties()

    def update_properties(self) -> None:
        """
        Meant to update trendLog properties like logBuffer,
        startTime, stopTime, recordCount, totalRecordCount, statusFlags, etc...
        """
        # startTime = self.data[0].timestamp
        # stopTime = self.data[-1].timestamp
        if not getattr(self.obj, "enable"):
            if getattr(self.obj, "recordCount") == 0:
                self.data = []  # empty it
            else:
                return  # disable....

        count = len(self.data)
        SequenceOfLogRecord = ListOf(LogRecord)()
        for each in self.data:
            logRecord = self.to_bacpypes_logrecord(each)
            SequenceOfLogRecord.append(logRecord)
        _props = {
            # "startTime": startTime,
            # "stopTime": stopTime,
            "logBuffer": SequenceOfLogRecord,
            "recordCount": Unsigned(count),
            "bufferSize": Unsigned(self.bufferSize),
            "enable": True,
            "stopWhenFull": False,
            "statusFlags": self.statusFlags,
            "totalRecordCount": Unsigned(self.data[-1].sequencenumber),
            "logInterval": Unsigned(self.data[-1].interval),
            "loggingType": LoggingType(0),
            "eventState": EventState(0),
            "reliability": Reliability(0),
        }
        for k, v in _props.items():
            setattr(self.obj, k, v)



#!/usr/bin/python
# type: ignore
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Simulate.py - simulate the value of controller I/O values
"""

from bacpypes3.app import Application

from ..app.asyncApp import BAC0Application
from .IOExceptions import (
    ApplicationNotStarted,
    NoResponseFromController,
    OutOfServiceNotSet,
    OutOfServiceSet,
)

# --- standard Python modules ---
# --- 3rd party modules ---
# --- this application's modules ---
from .Write import WriteProperty

# ------------------------------------------------------------------------------

VENDORS_REQUIRING_RELIABILITY_FOR_OoS = [5]
FORCE_RELIABILITY = True


class Simulation:
    """
    Global informations regarding simulation
    """

    async def sim(self, args, vendor_id: int = 0):
        """
        Simulate I/O points by setting the Out_Of_Service property, then doing a
        WriteProperty to the point's Present_Value.
        In the case of JCI controllers, we will also send a reliability "reliable" value
        to overcome the priority given to "communication error" over out of service.

        :param args: String with <addr> <type> <inst> <prop> <value> [ <indx> ] [ <priority> ]

        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app
        (
            address,
            obj_type,
            obj_inst,
            prop_id,
            value,
            priority,
            indx,
        ) = WriteProperty._parse_wp_args(args)

        if await self.is_out_of_service(args):
            await self._write(args)
            if vendor_id in VENDORS_REQUIRING_RELIABILITY_FOR_OoS or FORCE_RELIABILITY:
                await self.force_reliability(args)
        else:
            try:
                await self.out_of_service(args)
            except NoResponseFromController as e:
                self.log(
                    f"Failed to write to OutOfService property ({e})", level="warning"
                )

            try:
                if await self.is_out_of_service(args):
                    await self._write(
                        f"{address} {obj_type} {obj_inst} {prop_id} {value}"
                    )
                    if (
                        vendor_id in VENDORS_REQUIRING_RELIABILITY_FOR_OoS
                        or FORCE_RELIABILITY
                    ):
                        await self.force_reliability(args)
                else:
                    raise OutOfServiceNotSet()
            except NoResponseFromController as e:
                self.log(
                    f"Failed to write to OutOfService property ({e})", level="warning"
                )

    async def is_out_of_service(self, args):
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app
        (
            address,
            obj_type,
            obj_inst,
            prop_id,
            value,
            priority,
            indx,
        ) = WriteProperty._parse_wp_args(args)

        oos = await self.read(f"{address} {obj_type} {obj_inst} outOfService")

        return True if oos else False

    async def out_of_service(self, args):
        """
        Set the Out_Of_Service property so the Present_Value of an I/O may be written.

        :param args: String with <addr> <type> <inst> <prop> <value> [ <indx> ] [ <priority> ]

        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")
        (
            address,
            obj_type,
            obj_inst,
            prop_id,
            value,
            priority,
            indx,
        ) = WriteProperty._parse_wp_args(args)
        try:
            await self._write(f"{address} {obj_type} {obj_inst} outOfService True")
        except NoResponseFromController as e:
            self.log(f"Failed to write to OutOfService property ({e})", level="warning")

    async def force_reliability(self, args):
        """
        Set reliability property to NO_FAULT_DETECTED.
        Or else, in some cases, internal condition or previous reliability will prevent
        the value from being used internally.

        :param args: String with <addr> <type> <inst> <prop> <value> [ <indx> ] [ <priority> ]

        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")
        (
            address,
            obj_type,
            obj_inst,
            prop_id,
            value,
            priority,
            indx,
        ) = WriteProperty._parse_wp_args(args)
        try:
            await self._write(f"{address} {obj_type} {obj_inst} reliability 0")
        except NoResponseFromController as e:
            self.log(f"Failed to write to OutOfService property ({e})", level="warning")

    async def release(self, args):
        """
        Set the Out_Of_Service property to False - to release the I/O point back to
        the controller's control.

        :param args: String with <addr> <type> <inst>

        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        (
            address,
            obj_type,
            obj_inst,
            prop_id,
            value,
            priority,
            indx,
        ) = WriteProperty._parse_wp_args(args)
        try:
            await self._write(f"{address} {obj_type} {obj_inst} outOfService False")
        except NoResponseFromController as e:
            self.log(f"Failed to write to OutOfService property ({e})", level="warning")

        try:
            if await self.is_out_of_service(args) is True:
                raise OutOfServiceSet()
            else:
                pass  # Everything is ok"
        except NoResponseFromController as e:
            self.log(f"Failed to read OutOfService property ({e})", level="warning")



import typing as t
from collections import namedtuple

from bacpypes3.app import Application
from bacpypes3.basetypes import Date, DateTime, LogRecord, Polarity, Time, Unsigned
from bacpypes3.constructeddata import ArrayOf, ListOf
from bacpypes3.local.analog import (
    AnalogInputObject,
    AnalogOutputObject,
    AnalogValueObject,
)
from bacpypes3.local.binary import (
    BinaryInputObject,
    BinaryOutputObject,
    BinaryValueObject,
)
from bacpypes3.local.cov import COVIncrementCriteria
from bacpypes3.local.multistate import (
    MultiStateInputObject,
    MultiStateOutputObject,
    MultiStateValueObject,
)
from bacpypes3.primitivedata import CharacterString
from colorama import Fore

from BAC0.core.devices.local.trendLogs import LocalTrendLog

from ....scripts.Base import Base
from ...utils.notes import note_and_log
from .decorator import bacnet_properties, create, make_commandable, make_outOfService
from .object import (
    CharacterStringValueObject,
    DateTimeValueObject,
    DateValueObject,
    TrendLogObject,
)


@note_and_log
class ObjectFactory(object):
    """
    This is an exploration of a method to create local objects in BAC0 instance.

    First you need to know what bacpypes class of object you want to create ex. AnalogValueObject
    This class must be imported from bacpypes ex. from bacpypes.object import AnalogValueObject

    You must also define supplemental properties for the object in a dict.
    ex. :
        properties = {"outOfService" : False,
                "relinquishDefault" : 0,
                "units": "degreesCelsius",
                "highLimit": 98}

    Then you can use the factory to create your object
    ex. :

        av0 = ObjectFactory(AnalogValueObject, 1, 'av0', )

    """

    instances: t.Dict[str, t.Set] = {}

    definition = namedtuple(  # type: ignore[name-match]
        "Definition",
        "name, objectType, instance, properties, description, presentValue, is_commandable, relinquishDefault",
    )

    objects: t.Dict[str, t.Any] = {}
    # In the future... should think about a way to store relinquish default values because on a restart
    # those should be restored.

    def __init__(
        self,
        objectType,
        instance,
        objectName,
        properties=None,
        description="",
        presentValue=None,
        is_commandable=False,
        relinquishDefault=None,
    ):
        _localTrendLogDataType = properties.pop("trendLog_datatype", None)
        self._properties = ObjectFactory.default_properties(
            objectType, properties, is_commandable, relinquishDefault
        )
        # print(f"Obj {objectType} of type {type(objectType)}")
        if objectType is not TrendLogObject:
            pv_datatype = ObjectFactory.get_pv_datatype(objectType)
            self.log(f"pv datatype : {pv_datatype}", level="debug")

            def enforce_datatype(val, datatype):
                if not isinstance(val, datatype):
                    try:
                        val = datatype(val)
                    except TypeError as error:
                        raise TypeError(
                            f"Wrong datatype provided for value {val} using datatype {datatype} with target type of {pv_datatype} | {error}"
                        )
                return val

            if presentValue is not None:
                presentValue = enforce_datatype(presentValue, pv_datatype)
            if relinquishDefault is not None:
                relinquishDefault = enforce_datatype(relinquishDefault, pv_datatype)

        @bacnet_properties(self._properties)
        @make_commandable()
        def _create_commandable(
            objectType, instance, objectName, presentValue, description
        ):
            self._log.info(
                f"Making this object commandable : type:{objectType} id:{instance} name:{objectName} presentValue:{presentValue} description:{description}"
            )
            return create(objectType, instance, objectName, presentValue, description)

        @bacnet_properties(self._properties)
        @make_outOfService()
        def _create_outOfService(
            objectType, instance, objectName, presentValue, description
        ):
            return create(objectType, instance, objectName, presentValue, description)

        @bacnet_properties(self._properties)
        def _create(objectType, instance, objectName, presentValue, description):
            return create(objectType, instance, objectName, presentValue, description)

        objectName, instance = self.validate_name_and_instance(
            objectType, objectName, instance
        )

        if objectType in (AnalogInputObject, BinaryInputObject, MultiStateInputObject):
            self.objects[objectName] = _create_outOfService(
                objectType, instance, objectName, presentValue, description
            )
        elif is_commandable is True:
            self.objects[objectName] = _create_commandable(
                objectType, instance, objectName, presentValue, description
            )
        else:
            self.objects[objectName] = _create(
                objectType, instance, objectName, presentValue, description
            )
        if objectType is TrendLogObject:  # Add special internal object for trendlog
            self.objects[objectName]._local = LocalTrendLog(
                self.objects[objectName], datatype=_localTrendLogDataType
            )  # this will need to be fed by another process.
        else:
            self.objects[objectName]._cov_criteria = COVIncrementCriteria

    def validate_instance(self, objectType, instance):
        _warning = True
        try:
            _set = self.instances[objectType.__name__]
        except KeyError:
            _set = set()

        if not instance:
            instance = 0
            _warning = False

        if instance in _set:
            while instance in _set:
                instance += 1
            if _warning:
                self.log(
                    f"Instance alreaday taken, using {instance} instead",
                    level="warning",
                )

        _set.add(instance)
        self.instances[objectType.__name__] = _set
        return instance

    def validate_name_and_instance(self, objectType, objectName, instance):
        name_must_be_changed = False
        if objectName in self.objects.keys():
            name_must_be_changed = True
        instance = self.validate_instance(objectType, instance)
        if name_must_be_changed:
            objectName = f"{objectName}-{instance}"
            self.log(f"Name already taken, using {objectName} instead", level="warning")

        return (objectName, instance)

    @classmethod
    def from_dict(cls, definition):
        return cls(
            objectType=definition["objectType"],
            instance=definition["instance"],
            objectName=definition["name"],
            properties=definition["properties"],
            description=definition["description"],
            presentValue=definition["presentValue"],
            is_commandable=definition["is_commandable"],
        )

    @staticmethod
    def get_pv_datatype(objectType):
        return objectType.get_property_type("presentValue")

    @staticmethod
    def clear_objects():
        ObjectFactory.objects = {}
        ObjectFactory.instances = {}

    def add_objects_to_application(self, app):
        if isinstance(app, Base):
            app = app.this_application.app
        if not (isinstance(app, Application)):
            raise TypeError("Provide BAC0Application object or BAC0 Base instance")
        for k, v in self.objects.items():
            try:
                app.add_object(v)
                self.log(f"Adding {k} to application.", level="info")
            except RuntimeError:
                self._log.warning(
                    f"There is already an object named {k} in application."
                )

    def __repr__(self):
        return f"{self.objects}"

    @staticmethod
    def default_properties(
        objectType, properties, is_commandable=False, relinquishDefault=None
    ):
        _properties = properties or {}
        if "statusFlags" not in _properties.keys():
            _properties["statusFlags"] = [0, 0, 0, 0]
        if (
            "analog" in objectType.__name__.lower()
            and "units" not in _properties.keys()
        ):
            raise ValueError("Provide Engineering Units in properties")
        return _properties

    @staticmethod
    def relinquish_default_value(objectType, value):
        pv_datatype = ObjectFactory.get_pv_datatype(objectType)
        return pv_datatype(value)

    @staticmethod
    def inspect(bacnet_object):
        """
        A fun little snippet to inspect the properties of a BACnet object.
        """
        objectType = bacnet_object.__name__
        _repr = Fore.YELLOW + "*" * 100
        _repr += "\n| " + Fore.CYAN + f"{objectType}"
        _repr += Fore.YELLOW + "\n|" + "*" * 99
        _repr += "\n"
        _repr += (
            "| " + Fore.WHITE + f"Support COV : {bacnet_object._object_supports_cov}\n"
        )
        _repr += Fore.YELLOW + "|" + "=" * 99
        _repr += "\n\x1b[33m| {:30} \x1b[33m| {:26} \x1b[33m| {:8} \x1b[33m| {:8} \x1b[33m| {:8}".format(
            "IDENTIFIER", "DATACLASS", "OPTIONAL", "MUTABLE", "DEFAULT"
        )
        _repr += "\n|" + "=" * 99
        for prop in bacnet_object.properties:
            i, d, o, m, df = prop.__dict__.values()
            df = Fore.WHITE + df if df else Fore.WHITE + "None"
            o = Fore.GREEN + "Yes" if o else Fore.RED + "No"
            m = Fore.GREEN + "Yes" if m else Fore.RED + "No"
            _repr += (
                "\n"
                + Fore.WHITE
                + "\x1b[33m| \x1b[37m{:30} \x1b[33m| \x1b[37m{:26} \x1b[33m| {:13} \x1b[33m| {:13} \x1b[33m| {:13}".format(
                    i, d.__name__, o, m, df
                )
            )
        return _repr


"""
Those models are opiniated versions of BACnet objects.
They are meant to supply an easy mechanism to build BACnet objects
that will be served by the local device of BAC0.

You can indeed use this "method" and modify it to fill your needs.
"""


def _create(definition, **kwargs):
    _definition = {
        "name": None,
        "objectType": None,
        "instance": None,
        "properties": {},
        "description": None,
        "presentValue": None,
        "is_commandable": False,
    }
    _definition.update(definition)
    for k, v in kwargs.items():
        if k == "properties":
            for _k, _v in v.items():
                _definition["properties"][_k] = _v  # type: ignore[index]
        else:
            _definition[k] = v

    return ObjectFactory.from_dict(_definition)


def make_state_text(list_of_string: list[str]):
    _arr = ArrayOf(CharacterString)
    _lst = [CharacterString(each) for each in list_of_string]
    return _arr(_lst)


def set_default_if_not_provided(prop, default_prop, **kwargs):
    try:
        _prop = kwargs[prop]
    except KeyError:
        _prop = default_prop
    return _prop


def analog(**kwargs):
    # This is the general definition for an analog object
    # more information are provided in each of sub functions
    if "is_commandable" not in kwargs:
        kwargs["is_commandable"] = False
    definition = {
        "instance": 0,
        "description": "No description",
        "properties": {
            "units": "percent",
            # "eventState": EventState(),
            "covIncrement": 0.15,
        },
        "presentValue": 0,
    }
    return _create(definition, **kwargs)


def analog_input(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "AI", **kwargs)
    kwargs["objectType"] = AnalogInputObject
    kwargs["instance"] = set_default_if_not_provided("instance", 0, **kwargs)
    return analog(**kwargs)


def analog_output(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "AO", **kwargs)
    kwargs["objectType"] = AnalogOutputObject
    kwargs["instance"] = set_default_if_not_provided("instance", 0, **kwargs)
    return analog(**kwargs)


def analog_value(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "AV", **kwargs)
    kwargs["objectType"] = AnalogValueObject
    kwargs["instance"] = set_default_if_not_provided("instance", 0, **kwargs)
    return analog(**kwargs)


def binary(**kwargs):
    if "is_commandable" not in kwargs:
        kwargs["is_commandable"] = False
    definition = {
        "instance": 0,
        "description": "No description",
        "properties": {
            # "outOfService": Boolean(False),
            # "eventState": EventState()
        },
        "presentValue": "inactive",
    }
    return _create(definition, **kwargs)


def binary_input(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "BI", **kwargs)
    kwargs["objectType"] = BinaryInputObject
    try:
        kwargs["properties"].update({"polarity": Polarity("normal")})
    except KeyError:
        kwargs["properties"] = {"polarity": Polarity("normal")}
    return binary(**kwargs)


def binary_output(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "BO", **kwargs)
    kwargs["objectType"] = BinaryOutputObject
    try:
        kwargs["properties"].update({"polarity": Polarity("normal")})
    except KeyError:
        kwargs["properties"] = {"polarity": Polarity("normal")}
    return binary(**kwargs)


def binary_value(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "BV", **kwargs)
    kwargs["objectType"] = BinaryValueObject
    return binary(**kwargs)


def multistate(**kwargs):
    if "is_commandable" not in kwargs:
        kwargs["is_commandable"] = False
    default_states = make_state_text(["Off", "On"])
    definition = {
        "instance": 0,
        "properties": {
            # "eventState": EventState(),
            "stateText": default_states,
            # "numberOfStates": Unsigned(2),
        },
        "description": "No description",
        "presentValue": Unsigned(1),
        # "is_commandable": False,
        # "relinquishDefault": Unsigned(1),
    }
    return _create(definition, **kwargs)


def multistate_input(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "MSI", **kwargs)
    kwargs["objectType"] = MultiStateInputObject
    kwargs["relinquishDefault"] = Unsigned(1)

    return multistate(**kwargs)


def multistate_output(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "MSO", **kwargs)
    kwargs["objectType"] = MultiStateOutputObject
    kwargs["relinquishDefault"] = Unsigned(1)

    return multistate(**kwargs)


def multistate_value(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "MSV", **kwargs)
    kwargs["objectType"] = MultiStateValueObject
    kwargs["relinquishDefault"] = Unsigned(1)

    return multistate(**kwargs)


def temperature_input(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "TEMP_INPUT", **kwargs)
    kwargs["properties"] = {"units": "degreesCelsius"}
    return analog_input(**kwargs)


def temperature_value(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "TEMP_VALUE", **kwargs)
    kwargs["properties"] = {"units": "degreesCelsius"}
    return analog_value(**kwargs)


def humidity_input(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "HUM_INPUT", **kwargs)
    kwargs["properties"] = {"units": "percentRelativeHumidity"}
    return analog_input(**kwargs)


def humidity_value(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "HUM_VALUE", **kwargs)
    kwargs["properties"] = {"units": "percentRelativeHumidity"}
    return analog_value(**kwargs)


def character_string(**kwargs):
    definition = {
        "name": "CS_VALUE",
        "objectType": CharacterStringValueObject,
        "instance": 0,
        "description": "No description",
        "presentValue": "",
        "properties": {},
        "is_commandable": False,
        "relinquishDefault": "inactive",
    }
    return _create(definition, **kwargs)


def date_value(**kwargs):
    definition = {
        "name": "DATE_VALUE",
        "objectType": DateValueObject,
        "instance": 0,
        "description": "No description",
        "presentValue": Date().now(),
        "properties": {},
        "is_commandable": False,
        # "relinquishDefault": "inactive",
    }
    return _create(definition, **kwargs)


def datetime_value(**kwargs):
    definition = {
        "name": "DATETIME_VALUE",
        "objectType": DateTimeValueObject,
        "instance": 0,
        "description": "No description",
        "presentValue": DateTime(date=Date().now(), time=Time().now()),
        "properties": {},
        "is_commandable": False,
    }
    return _create(definition, **kwargs)


def trendlog(**kwargs):
    definition = {
        "name": "TREND_LOG",
        "objectType": TrendLogObject,
        "instance": 0,
        "description": "No description",
        "properties": {
            "enable": True,
            "logBuffer": ListOf(LogRecord)([]),
            # "logDeviceObjectProperty": DeviceObjectPropertyReference(
            #    objectIdentifier=ObjectIdentifier("trendLog", 0),
            # ),
            "trendLog_datatype": "realValue",
        },
    }
    return _create(definition, **kwargs)
